#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
opticiq.gen
===========
For generating pattern images used in image quality testing.

See ChartGen_pdf, ChartGen_img, ChartGen_gds

Created on Sat Feb 25 15:13:43 2023

@author: chris
"""

from .chart import *
from .extra import *
from .gauss import *
from .model import *
from .star import *
