#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
opticiq.opt subpackage contains very simple optical models such as FFT-->PSF.
These can be useful helpers, e.g. forward models can be good for code testing,
data fitting algoriths, or Monte Carlo methods.

Created on Sat Feb  1 11:39:20 2025

@author: chris
"""

from .fft import *
